/*
 * Copyright  2004 Mass Dosage
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 */
package za.co.massdosage.ant;

import java.io.BufferedInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.security.InvalidKeyException;
import java.security.Key;
import java.security.NoSuchAlgorithmException;

import javax.crypto.Cipher;
import javax.crypto.CipherOutputStream;
import javax.crypto.KeyGenerator;
import javax.crypto.NoSuchPaddingException;

import org.apache.tools.ant.BuildException;
import org.apache.tools.ant.Project;
import org.apache.tools.ant.Task;

/**
 * Ant task for generating symmetric encryption keys and encrypting and decrypting files.
 * 
 * @author mass
 */
public class Crypter extends Task {
    
    /**
     * The default cipher transformation used for encryption or decryption.
     */
    public static final String DEFAULT_CIPHER_TRANSFORMATION = "Blowfish/ECB/PKCS5Padding";
    
    /**
     * The default algorithm used if generating a key. 
     */
    public static final String DEFAULT_KEY_ALGORITHM = "Blowfish";

    private File keyFile;
  	private File inputFile;
  	private File outputFile;
  	private boolean encrypt = true;
  	private boolean generateKey = false;
  	private String cipherTransformation = DEFAULT_CIPHER_TRANSFORMATION;
    private String keyAlgorithm = DEFAULT_KEY_ALGORITHM; 
    
    /**
     * Logs an informational message. This method is required so that this task can be used outside of ant.
     * 
     * @param message Message to log.
     */
    private void logInfo(String message) {
        if (this.getProject()!=null) { //we are running in ant, so use ant log
            this.log(message, Project.MSG_INFO);
        } else { //we are running outside of ant, log to System.out
            System.out.println(message);
        }
    }
   
    /**
     * Reads the contents of the key file and converts this into a <code>Key</code>.
     * 
     * @return The <code>Key</code> object.
     * @throws BuildException If the contents of the key file cannot be read. 
     */
  	private Key readKey() throws BuildException {
  	    if (this.keyFile==null) {
	        throw new BuildException("No 'keyFile' specified, cannot continue.");
	    }
  	    Key key;
  	    try {
  	        ObjectInputStream in;
  	        in = new ObjectInputStream(new BufferedInputStream(new FileInputStream(keyFile)));
  	        key = (Key)in.readObject();
  	        in.close();
  	    } catch (FileNotFoundException e) {
  	        throw new BuildException("Could not find key file " + this.keyFile.getAbsolutePath(), e);
  	    } catch (IOException e) {
  	        throw new BuildException("Error reading key from file " + this.keyFile.getAbsolutePath(), e);
  	    } catch (ClassNotFoundException e) {
  	        throw new BuildException("Class of serialised key not found " + e.getMessage(), e);
  	    }
  	    return key;
  	}
  	
  	/**
  	 * Initialises a <code>Cipher</code> in the mode set in the ant task (encrypt or decrypt) with the passed 
  	 * <code>Key</code>.
  	 *  
  	 * @param key The <code>Key</code> which the <code>Cipher</code> will use for encryption or decryption.
  	 * @return The initialised <code>Cipher</code>.
  	 * @throws BuildException If an error occurs initialising the cipher.
  	 */
  	private Cipher initialiseCipher(Key key) throws BuildException {
  	  Cipher cipher = null;
	    try {
	        cipher = Cipher.getInstance(this.cipherTransformation);
	        if (encrypt) {
	            cipher.init(Cipher.ENCRYPT_MODE, key);
	            this.logInfo("Initialised cipher to perform encryption using " + this.cipherTransformation);
	        } else {
	            cipher.init(Cipher.DECRYPT_MODE, key);
	            this.logInfo("Initialised cipher to perform decryption using " + this.cipherTransformation);
	        }
	    } catch (NoSuchAlgorithmException e) {
	        throw new BuildException("Cipher transformation algorithm [" + this.cipherTransformation + 
	                "] not supported", e);
	    } catch (NoSuchPaddingException e) {
	        throw new BuildException("Cipher padding scheme not supported", e);
	    }  catch (InvalidKeyException e) {
	        throw new BuildException("Invalid key for cipher", e);
	    }
  	    return cipher;
  	}
  	
  	/**
  	 * Performs the encryption/decryption according to the state of the passed <code>Cipher</code>, using the 
  	 * input and output files set in the ant task.
  	 * 
  	 * @param cipher An initialised <code>Cipher</code> to use for the encryption/decryption.
  	 * @throws BuildException If the input or output files cannot be found, read, or written to; or if an error 
  	 * occurs performing the encryption/decryption.
  	 */
  	private void crypt(Cipher cipher) throws BuildException {
  	  FileInputStream in = null;
	    try {
	        in = new FileInputStream(this.inputFile);
	    } catch (FileNotFoundException e) {
	        throw new BuildException("Could not find input file " + this.inputFile, e);
	    }
	    FileOutputStream fileout = null;
	    try {
	        fileout = new FileOutputStream(this.outputFile);
	    } catch (FileNotFoundException e) {
	        throw new BuildException("Invalid output file " + this.outputFile, e);
	    }
	    
	    CipherOutputStream out = new CipherOutputStream(fileout, cipher);
	    byte[] buffer = new byte[8192];
	    int length;
	    try {
	        while ((length = in.read(buffer)) != -1) {
	            out.write(buffer,0,length);
	        }
	        in.close();
	        out.close();
	    } catch (IOException e) {
	        throw new BuildException("Error writing output file " + this.outputFile, e);
	    }
	    this.logInfo("Performed cryptographic transformation on " + this.inputFile.getAbsolutePath() + " to " + 
	            this.outputFile.getAbsolutePath());
  	}
  	
  	/**
  	 * Generates a <code>Key</code> and serialises it to disk.
  	 * 
  	 * @throws BuildException If an invalid key algorithm was specified in the task or something goes wrong generating 
  	 * they key or serialising it.
  	 */
  	private void generateKey() throws BuildException {
  	    if (this.keyFile==null) {
  	        throw new BuildException("No 'keyFile' specified, cannot generate key");
  	    }
  	    
 	    if (this.keyFile.getParentFile()!=null && !this.keyFile.getParentFile().exists()) {
  	        this.keyFile.getParentFile().mkdirs(); //create parent folder if necessary
  	    }
  	    
  	    KeyGenerator keyGenerator;
        try {
            keyGenerator = KeyGenerator.getInstance(this.keyAlgorithm);
        } catch (NoSuchAlgorithmException e) {
            throw (new BuildException("Invalid key algorithm", e));
        }
        Key key = keyGenerator.generateKey();
  	    ObjectOutputStream keyOutputStream;
        try {
            keyOutputStream = new ObjectOutputStream(new FileOutputStream(this.keyFile));
            keyOutputStream.writeObject(key);
        } catch (FileNotFoundException e) {
            throw new BuildException("Invalid key file " + this.keyFile.getAbsolutePath(), e);
        } catch (IOException e) {
            throw new BuildException("Error writing key file " + this.keyFile.getAbsolutePath(), e);
        }
        this.logInfo("Generated a " + this.keyAlgorithm + " key to " + this.keyFile.getAbsolutePath());
  	}
  	
  	
  	/**
  	 * Called by the project to perform the encryption or decryption using the parameters set in the task.
  	 *  
  	 * @throws BuildException If something goes wrong executing this task.
  	 */
  	public void execute() throws BuildException {
  	    if (this.generateKey) {
  	        this.generateKey();
  	    }
  	    if (!(this.inputFile==null && this.outputFile==null)) { //if input or output files specified, need to attmept enc/dec
  	        Key key = this.readKey();
  	        Cipher cipher = this.initialiseCipher(key);
  	        this.crypt(cipher);
  	    }
  	}
  	
  	/**
  	 * Sets the algorithm used to generate a <code>Key</code>. If this is not set, then the default value specified by 
  	 * <code>DEFAULT_KEY_ALGORITHM</code> will be used.
  	 * 
  	 * @param keyAlgorithm The standard name of the requested key algorithm. 
  	 */
  	public void setKeyAlgorithm(String keyAlgorithm) {
  	    this.keyAlgorithm = keyAlgorithm;
  	}

  	/**
  	 * Sets the location of the <code>File</code> containing the <code>Key</code> to be used for encryption/decryption.
  	 * 
  	 * @param keyFile The location of the key file.
  	 */
  	public void setKeyFile(File keyFile) {
    	this.keyFile = keyFile;
  	}
  	
  	/**
  	 * Sets the location of the input <code>File</code> that is to be encrypted/decrypted.
  	 * 
  	 * @param inputFile The location of the input file. 
  	 */
    public void setInputFile(File inputFile) {
    	this.inputFile = inputFile;
  	}

    /**
  	 * Sets the location of the output <code>File</code> that is the results of the encryption/decryption.
  	 * 
  	 * @param outputFile The location of the output file. 
  	 */
  	public void setOutputFile(File outputFile) {
    	this.outputFile = outputFile;
  	}

  	/**
  	 * Sets the mode that is used to determine whether encryption or decryption will be performed. If the value 
  	 * "true" is passed to this method then encryption will be performed, if the value "false" is passed then 
  	 * decryption will be performed. If this value is not set, encryption will be performed by default.
  	 * 
  	 * @param encrypt Whether to perform encryption or decryption.
  	 */
	public void setEncrypt(boolean encrypt) {
        this.encrypt = encrypt;
    }
	
	/**
	 * Determines whether to generate a key file. If this value is not set, a key file will not be generated by 
	 * default.
	 *  
	 * @param generateKey Whether to generate a key or not.
	 */
	public void setGenerateKey(boolean generateKey) {
	    this.generateKey = generateKey;
	}

	/**
	 * Sets the cipher transformation that will be used to perform the encryption/decryption. If this is not set, then 
	 * the default value specified by <code>DEFAULT_CIPHER_TRANSFORMATION</code> will be used.
	 * 
	 * @param transformation The name of the transformation, for example <i>Blowfish/ECB/PKCS5Padding</i>. 
	 */
  	public void setCipherTransformation(String transformation) {
    	this.cipherTransformation = transformation;
  	}

}

